package net.pingex.dcf.modularity;

import net.pingex.dcf.core.Configuration;
import org.apache.commons.io.FilenameUtils;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import java.io.IOException;
import java.net.URL;
import java.net.URLClassLoader;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.HashMap;
import java.util.Map;
import java.util.jar.Attributes;
import java.util.jar.JarFile;

/**
 * Class that load the plugins, a very powerful thingy.
 */
public class PluginLoader
{
	/**
	 * Singleton instance
	 */
	private static PluginLoader INSTANCE = new PluginLoader();

	/**
	 * Logger
	 */
	private static Logger LOGGER = LogManager.getLogger(PluginLoader.class);

	/**
	 * All plugins
	 */
	Map<String, PluginWrapper> plugins;

	/**
	 * Gives the singleton instance
	 * @return The singleton instance
	 */
	public static PluginLoader getInstance()
	{
		return INSTANCE;
	}

	/**
	 * Enforce singleton with a private constructor.
	 */
	private PluginLoader()
	{
		plugins = new HashMap<>();
	}

	/**
	 * Discover all plugins contained in .jar files in the plugins directory.
	 */
	public void discoverPluginsDirectory()
	{
		LOGGER.debug("Discovering all IPlugin classes in plugins directory.");

		try
		{
			Files.walk(Paths.get(Configuration.PLUGINS_DIR))
					.filter(Files::isRegularFile)
					.filter(path -> FilenameUtils.getExtension(path.toString()).equals("jar"))
					.forEach(path -> {
						try
						{
							loadJarFile(path);
						}
						catch(IOException e)
						{
							LOGGER.catching(e);
						}
					});
		}
		catch(IOException e)
		{
			LOGGER.error("IO error while discovering plugins.", e);
		}
	}

	/**
	 * Look in a .jar file for IPlugins.
	 */
	public void loadJarFile(Path target) throws IOException
	{
		LOGGER.debug("Looking up {} for plugins.", target.getFileName());

		// Get DCF-Plugins manifest attribute
		Attributes jarAttributes = new JarFile(target.toFile()).getManifest().getMainAttributes();
		LOGGER.trace("DCF-Plugins attribute: {}", jarAttributes.getValue("DCF-Plugins"));
		if(jarAttributes.getValue("DCF-Plugins") == null)
		{
			LOGGER.debug("Jar file has no `DCF-Plugins` attribute in its manifest. Ignoring archive.");
			return;
		}

		String[] pluginsToLoad = jarAttributes.getValue("DCF-Plugins").split(" ");
		LOGGER.debug("Manifest returned {} plugins to load.", pluginsToLoad.length);
		URLClassLoader classLoader = URLClassLoader.newInstance(new URL[]{target.toUri().toURL()});

		for(String item : pluginsToLoad)
		{
			LOGGER.debug("Attempting to load {}.", item);

			Class<?> clazz;
			try
			{
				clazz = Class.forName(item, true, classLoader);
			}
			catch(ClassNotFoundException e) // Class with that name doesn't exist.
			{
				LOGGER.debug("No class with name {} was found.", item);
				return;
			}

			// Class doesn't implement IPlugin, or doesn't have @Plugin annotation.
			if(!clazz.isAnnotationPresent(Plugin.class) || !IPlugin.class.isAssignableFrom(clazz))
			{
				LOGGER.debug("Class does not implement IPlugin or doesn't gave @Plugin annotation.");
				return;
			}

			LOGGER.debug("Plugin {} {} preloaded.", clazz.getAnnotation(Plugin.class).id(), clazz.getAnnotation(Plugin.class).version());
			plugins.put(clazz.getAnnotation(Plugin.class).id(), new PluginWrapper(clazz.asSubclass(IPlugin.class), classLoader));
		}
	}
}